import random, re, string, sys

from . import debug, multidict

# TODO: Formatting directives ('{line break}', '{no line break}', etc.)
# TODO: 'the number of', etc.?

LINE_WIDTH = 80

STOPWORDS = [
  'the',
  'a',
  'an',
  'of',
  'in',
  'at',
]

def first(thing):
  for it in thing:
    return it

def np(object):
  if hasattr(object, "noun_phrase"):
    return object.noun_phrase()
  return str(object)

def titlecase(val):
  result = re.sub(
    r"(\A|[^'])\b([a-z])",
    lambda m: m.group(1) + m.group(2).upper(),
    val,
  )
  for sw in STOPWORDS:
    result = re.sub(
      '(?!\\A|["\'])\\b' + sw.title() + r"\b",
      lambda m: m.group(0).lower(),
      result,
    )
  return result

# Pronoun classifications:
PRONOUNS = multidict.multidict()
# By number:
PRONOUNS.map_all(
  'singular',
    "I", "me", "myself", "mine", "my",
    "you", "yourself", "yours", "your",
    "ey", "em", "eirself", "eirs", "eir",
    "he", "him", "himself", "his",
    "she", "her", "herself", "hers",
    "it", "itself", "its",
    "one", "oneself", "one's",
)
PRONOUNS.map_all(
  'plural',
    "we", "us", "ourselves", "ours", "our",
    "you", "yourselves", "yours", "your",
    "they", "them", "themselves", "theirs", "their",
)

# By function/case:
PRONOUNS.map_all(
  'subjective',
  "I", "we", "you", "ey", "he", "she", "it", "one", "they"
)
PRONOUNS.map_all(
  'objective',
  "me", "us", "you", "em", "him", "her", "it", "one", "them"
)
PRONOUNS.map_all(
  'reflexive',
    "myself", "ourselves", "yourself", "yourselves", "eirself", "himself",
    "herself", "itself", "oneself", "themselves"
)
PRONOUNS.map_all(
  'possessive',
  "mine", "ours", "yours", "eirs", "his", "hers", "its", "theirs"
)
PRONOUNS.map_all(
  'determinant',
    "my", "our", "your", "eir", "his", "her", "its", "one's", "their"
)

# By person:
PRONOUNS.map_all(
  'first',
    "I", "me", "myself", "mine", "my",
    "we", "us", "ourselves", "ours", "our",
)
PRONOUNS.map_all(
  'second',
    "you", "yourself", "yours", "your",
)
PRONOUNS.map_all(
  'third',
    "ey", "em", "eirself", "eirs", "eir",
    "he", "him", "himself", "his",
    "she", "her", "herself", "hers",
    "it", "itself", "its",
    "one", "oneself", "one's",
    "they", "them", "themselves", "theirs", "their",
)

# By gender:
PRONOUNS.map_all('neutral', "ey", "em", "eirself", "eirs", "eir")
PRONOUNS.map_all('masculine', "he", "him", "himself", "his")
PRONOUNS.map_all('feminine', "she", "her", "herself", "hers")
PRONOUNS.map_all('neuter', "it", "itself", "its")

PRONOUN_DEFAULTS = {
  'number': 'singular',
  'case': 'objective',
  'person': 'third',
  'gender': 'neutral',
}

SHORTCUT_PRONOUNS = (
  "I",
  "me",
  "myself",
  "mine",
  "my",
  "thou",
  "thee",
  "thyself",
  "thine",
  "thy",
  "ey",
  "em",
  "eirself",
  "eirs",
  "eir",
)

def pronoun(**keys):
  indices = dict(PRONOUN_DEFAULTS)
  indices.update(keys)
  number = indices['number']
  case = indices['case']
  person = indices['person']
  gender = indices['gender']
  if number not in PRONOUNS:
    number = PRONOUN_DEFAULTS['number']
  if case not in PRONOUNS:
    case = PRONOUN_DEFAULTS['case']
  if person not in PRONOUNS:
    person = PRONOUN_DEFAULTS['person']
  if gender not in PRONOUNS:
    gender = PRONOUN_DEFAULTS['gender']
  candidates = PRONOUNS[number] & PRONOUNS[case] & PRONOUNS[person]
  if len(candidates) == 1:
    return first(candidates)
  else:
    candidates = candidates & PRONOUNS[gender]
  if not candidates:
    debug.debug("Formatting", "No pronoun found for keys: {}".format(indices))
    return "it"
  return first(candidates)

def pronoun_for(object, case="objective", person="third"):
  number = "singular"
  if is_plural(object):
    number = "plural"
  if hasattr(object, "gender"):
    gender = object.gender or "neutral"
    if gender not in PRONOUNS:
      gender = "neutral"
  else:
    gender = "neuter"
  return pronoun(
    number=number,
    case=case,
    person=person,
    gender=gender,
  )

CARDINALS = {
  0: "zero",
  1: "one",
  2: "two",
  3: "three",
  4: "four",
  5: "five",
  6: "six",
  7: "seven",
  8: "eight",
  9: "nine",
  10: "ten",
  11: "eleven",
  12: "twelve",
  13: "thirteen",
  14: "fourteen",
  15: "fifteen",
  16: "sixteen",
  17: "seventeen",
  18: "eighteen",
  19: "nineteen",
  20: "twenty",
  30: "thirty",
  40: "forty",
  50: "fifty",
  60: "sixty",
  70: "seventy",
  80: "eighty",
  90: "ninety",
  100: "hundred",
  1000: "thousand",
  1000000: "million",
  1000000000: "billion",
  1000000000000: "trillion",
  1000000000000000: "quadrillion",
  1000000000000000000: "quintillion",
}

ORDINALS = {
  0: "zeroth",
  1: "first",
  2: "second",
  3: "third",
  4: "fourth",
  5: "fifth",
  6: "sixth",
  7: "seventh",
  8: "eighth",
  9: "ninth",
  10: "tenth",
  11: "eleventh",
  12: "twelfth",
  13: "thirteenth",
  14: "fourteenth",
  15: "fifteenth",
  16: "sixteenth",
  17: "seventeenth",
  18: "eighteenth",
  19: "nineteenth",
  20: "twentieth",
  30: "thirtieth",
  40: "fortieth",
  50: "fiftieth",
  60: "sixtieth",
  70: "seventieth",
  80: "eightieth",
  90: "ninetieth",
  100: "hundredth",
  1000: "thousandth",
  1000000: "millionth",
  1000000000: "billionth",
  1000000000000: "trillionth",
  1000000000000000: "quadrillionth",
  1000000000000000000: "quintillionth",
}

def cardinal(value):
  '''
  Takes a number and returns a string of words describing it. Starts returning
  strange results with "one thousand quintillion".
  '''
  rep = ""
  if value == 0:
    return CARDINALS[value]
  if value < 0:
    rep += "negative "
    value = -value
  while value:
    largest = max(k for k in CARDINALS if k <= value)
    upper = value//largest
    if largest >= 100:
      rep += cardinal(upper) + ' '
    elif upper > 1:
      debug.debug(
        "Formatting",
        "Warning: bad upper value {} (largest = {}, value = {})".format(
          upper, largest, value
        )
      )
    rep += CARDINALS[largest]
    if (20 <= largest <= 90):
      rep += '-'
    else:
      rep += ' '
    value -= largest*upper
  return rep[:-1]

def ordinal(value, use_and=True):
  '''
  Same as cardinal, but returns the ordinal form of the number. If 'use_and' is
  True, the word 'and' will be used before the final part of the number where
  appropriate (compare "one hundred twenty-fourth" with "one hundred and
  twenty-fourth")
  '''
  original = value
  rep = ""
  if value == 0:
    return ORDINALS[value]
  if value < 0:
    rep += "negative "
    value = -value
  while value:
    largest = max(k for k in CARDINALS if k <= value)
    smallest = max(k for k in CARDINALS if (k > 0 and value%k == 0))
    upper = value//largest
    if largest >= 100:
      rep += cardinal(upper) + ' '
    elif upper > 1:
      debug.debug(
        "Formatting",
        "Warning: bad upper value {} (largest = {}, value = {})".format(
          upper, largest, value
        )
      )
    if largest == smallest:
      rep += ORDINALS[smallest]
    else:
      rep += CARDINALS[largest]
    if (20 <= largest <= 90):
      rep += '-'
    else:
      rep += ' '
    value -= largest*upper
  if original >= 100 and use_and:
    where = rep[:-1].rindex(' ')
    rep = rep[:where] + " and" + rep[where:]
  return rep[:-1]

def list_position(value, use_and=None):
  '''
  Like ordinal, but handles list indices, so 0 -> first, 1 -> second, and
  additionally -1 -> last, -2 -> second-to-last, etc. If use_and is not given,
  it will default to True for positive indices and False for negative ones.
  '''
  rep = ""
  if value == -1:
    return 'last'
  elif value < 0:
    if use_and is None:
      return ordinal(-value, False) + "-to-last"
    else:
      return ordinal(-value, use_and) + "-to-last"
  else:
    if use_and is None:
      return ordinal(value + 1, True)
    else:
      return ordinal(value + 1, use_and)

def listfmt(value):
  '''
  Returns a readable formatted list.
  '''
  if not value:
    return ''
  elif len(value) == 1:
    return np(value[0])
  elif len(value) == 2:
    return np(value[0]) + " and " + np(value[1])
  else:
    return ', '.join(np(v) for v in value[:-1]) + " and " + np(value[-1])

def number_of(thing):
  length = None
  if isinstance(thing, str):
    return 1
  if hasattr(thing, "__len__"):
    if len(thing) > 1 or len(thing) is 0:
      length = len(thing)
    elif len(thing) is 1:
      for t in thing:
        thing = t
        break
  elif hasattr(thing, "__iter__"):
    count = 0
    for _ in thing:
      count += 1
    if count > 1 or count is 0:
      length = count
    else:
      for t in thing:
        thing = t
        break
  if length is None:
    if hasattr(thing, "plural") and thing.plural:
      length = 2
    else:
      length = 1

  return length

def is_plural(thing):
  return number_of(thing) > 1

def sentence(*stuff):
  string = ''.join(str(thing) for thing in stuff)
  if not string:
    return ''
  if string[-1] not in ('\n', '.'):
    string += '.'
  string = string[0].upper() + string[1:]
  return string

def reflow(text, preindent = 0, postindent = 0):
  text = re.sub(r"[ \n\t]+([,;:.])", r"\1", text)
  words = text.strip().split()
  if not words:
    return ''
  lines = [' '*preindent]
  while words:
    if len(lines[-1]) + len(words[0]) + 1 < LINE_WIDTH:
      lines[-1] += ' ' + words[0]
      words = words[1:]
    elif len(words[0]) + 1 >= LINE_WIDTH:
      lines.append(' '*postindent + words[0])
      words = words[1:]
    else:
      lines.append(' '*postindent)
  return '\n'.join(lines)

def paragraph(text, indent=2):
  if indent >= 0:
    pre = indent
    post = 0
  else:
    pre = 0
    post = -indent
  result = reflow(text, preindent=pre, postindent=post)
  if result:
    return result + '\n\n'
  return ""

utils = [
  np,
  titlecase,
  pronoun,
  pronoun_for,
  cardinal,
  ordinal,
  list_position,
  listfmt,
  number_of,
  is_plural,
  sentence,
  reflow,
  paragraph,
]


class templateFormatter(string.Formatter):
  '''
  An extended formatter that allows template constructions, and makes some
  extra formatting methods accessible. The following template constructions are
  implemented:
    {if <condition>}...{elif condition}...{else}...{end}
    {one of}...{or}...{or}...{at random|indexed by <value>}
    {is/are <anything or list>}
  Also, a list of various methods of the Anything class (most of the
  zero-argument methods) can be called using a prefix notation: {the <object>}
  is replaced with the result of <object>.the(), etc. A list of methods that
  can be invoked in this manner is given in the 'premethods' field. Note that
  this formatter has a single special formatting argument, called
  "_CODE_PREP_". If given, this function is removed from the normal keyword
  arguments list, and is run on all code segments prior to evaluation.
  '''
  premethods = [
    "A",
    "a",
    "An",
    "an",
    "The",
    "the",
    "At the",
    "at the",
    "At",
    "at",
    "I",
    "me",
    "myself",
    "mine",
    "my",
    "thou",
    "thee",
    "thyself",
    "thine",
    "thy",
    "ey",
    "em",
    "eirself",
    "eirs",
    "eir",
    "title",
    "a title",
    "the title",
    "describe",
    "capitalize",
  ]
  def vformat(self, format_string, args, kwargs):
    mode = "normal"
    result = ""
    stack = []
    referent = None

    if "_CODE_PREP_" in kwargs:
      prep = kwargs["_CODE_PREP_"]
      del kwargs["_CODE_PREP_"]
    else:
      prep = lambda x: x

    def add_partial(value):
      nonlocal result, stack
      if value is None:
        return
      if stack:
        if stack[-1]['mode'] == "conditional":
          stack[-1]['partial'] += value
        elif stack[-1]['mode'] == "options":
          stack[-1]['options'][-1] += value
      else:
        result += value

    def pop_frame():
      nonlocal stack
      add_partial(stack.pop()['result'])

    for literal, field, spec, conversion in self.parse(format_string):
      # Add whatever literal text was given:
      add_partial(literal)

      # Update the referent:
      kwargs['referent'] = referent

      # Determine the mode:
      mode = "normal"
      if stack:
        frame = stack[-1]
        mode = frame['mode']

      if not field:
        continue

      # Check for the start of a new stack entry or for a special field (these
      # are mode-independent field values):
      if field[:3] == "if " and field[3:]:
        condition = field[3:]
        try:
          val = bool(eval(prep(condition), globals(), kwargs))
        except:
          debug.debug("Formatting", "Condition crashed.")
          if debug.domainactive("Formatting"):
            sys.excepthook(*sys.exc_info())
          val = False
        stack.append({
          'mode': "conditional",
          'result': "",
          'partial': "",
          'condition': val,
        })

      elif field == "one of":
        stack.append({
          'mode': "options",
          'result': "",
          'options': [""],
          'condition': False,
        })

      elif any(field[:len(pm) + 1] == (pm + ' ') and field[len(pm) + 1:]
               for pm in self.premethods):
        method = None
        for pm in self.premethods:
          if field[:len(pm) + 1] == pm + ' ':
            method = pm
            break
        try:
          referent = eval(prep(field[len(method) + 1:]), globals(), kwargs)
        except:
          debug.debug("Formatting", "Premethod '{}' target unavailable.".format(
            method
          ))
          if debug.domainactive("Formatting"):
            sys.excepthook(*sys.exc_info())
          referent = field[len(method) + 1:]
        try:
          evald = getattr(referent, method.replace(' ', '_'))()
        except:
          debug.debug(
            "Formatting",
            "Premethod '{}' unavailable for given target.".format(method)
          )
          evald = np(referent)
        add_partial(evald)

      elif field.split('[')[0] in SHORTCUT_PRONOUNS:
        pronoun = field.split('[')[0]
        if field.split('[')[1:]:
          obj = '['.join(field.split('[')[1:])
          if obj[-1] == ']':
            obj = obj[:-1]
          else:
            debug.debug("Formatting", "Pronoun object not bracketed.")
          try:
            referent = eval(prep(obj), globals(), kwargs)
          except:
            debug.debug("Formatting", "Bad pronoun object.")
            if debug.domainactive("Formatting"):
              sys.excepthook(*sys.exc_info())
            referent = None
        if referent:
          try:
            pronoun = getattr(referent, pronoun)()
          except:
            debug.debug(
              "Formatting",
              "Pronoun '{}' unavailable from referent.".format(pronoun)
            )
            if debug.domainactive("Formatting"):
              sys.excepthook(*sys.exc_info())
        add_partial(pronoun)

      elif field[0] == '/' and\
           (field[-1] == '/' and '/' in field[1:-1]) or\
           (field[-1] == ']' and '/[' in field[1:-1] and\
            '/' in field[1:field.index('/[')]):
        bits = field.split('/')[1:]
        sg = bits[0]
        pl = bits[1]
        obj = '/'.join(bits[2:])
        if obj:
          if obj[0] == '[' and obj[-1] == ']':
            obj = obj[1:-1]
          else:
            debug.debug("Formatting", "Pluralization object not bracketed.")
          try:
            referent = eval(prep(obj), globals(), kwargs)
          except:
            debug.debug("Formatting", "Bad pluralization object.")
            if debug.domainactive("Formatting"):
              sys.excepthook(*sys.exc_info())
            referent = None
        if referent:
          number = number_of(referent)
        else:
          debug.debug("Formatting", "Missing referent for pluralization.")
          number = 1
        # If number is 0, don't add anything...
        if number == 1:
          add_partial(sg)
        elif number > 1:
          add_partial(pl)

      # Handle mode-specific field values:
      elif mode == "conditional" and field[:5] == "elif " and field[5:]:
        if frame['condition'] == True:
          frame['result'] = frame['partial']
          frame['condition'] = "done"
        elif frame['condition'] == False:
          condition = field[5:]
          try:
            val = bool(eval(prep(condition), globals(), kwargs))
          except:
            debug.debug("Formatting", "Condition crashed.")
            if debug.domainactive("Formatting"):
              sys.excepthook(*sys.exc_info())
            val = False
          frame['condition'] = val
        frame['partial'] = ""

      elif mode == "conditional" and field == "else":
        if frame['condition'] == True:
          frame['result'] = frame['partial']
          frame['condition'] = "done"
        elif frame['condition'] == False:
          frame['condition'] = True
        frame['partial'] = ""

      elif mode == "conditional" and field == "end":
        if frame['condition'] == True:
          frame['result'] = frame['partial']
        pop_frame()

      elif mode == "options" and field == "or":
        frame['options'].append("")

      elif mode == "options" and field in ("randomly", "end"):
        frame['result'] = random.choice(frame['options'])
        pop_frame()

      elif mode == "options" and field[:11] == "indexed by " and field[11:]:
        try:
          index = int(eval(prep(field[11:]), globals(), kwargs))
          frame['result'] = frame['options'][index]
        except:
          debug.debug("Formatting", "Invalid index (using 0).")
          if debug.domainactive("Formatting"):
            sys.excepthook(*sys.exc_info())
          frame['result'] = frame['options'][0]
        pop_frame()

      else:
        # For normal fields, compute normal formatting operations and add them
        # to the local result:
        try:
          referent, used = self.get_field(field, args, kwargs)
          obj = self.convert_field(referent, conversion)
          format_spec = self.vformat(spec, args, kwargs)
          add_partial(self.format_field(obj, format_spec))
        except KeyError:
          debug.debug(
            "Formatting",
            "Unable to look up field '{}'.".format(field)
          )
          add_partial(field)

    return result


  def format_field(self, value, spec):
    '''
    Adds some custom formatting methods for various types, accessible via
    custom format_spec values. This implements things like {number:ordinal}
    using the various formatting functions defined in this module.
    '''
    if type(value) == int:
      if spec == 'cardinal':
        return cardinal(value)
      elif spec == 'ordinal':
        return ordinal(value)
      elif spec == 'list-position':
        return ordinal(value + 1)
    elif spec == 'list':
      try:
        return listfmt(value)
      except:
        return format(value, spec)
    return format(value, spec)

formatter = templateFormatter()

def sp_format(fmt_string, *args, **kwargs):
  return formatter.format(fmt_string, *args, **kwargs)

utils.append(sp_format)
